#version 440 core

out vec4 FragColor;

#define NR_POINT_LIGHTS 1

struct DirLight {
    vec3 direction;
	
    vec3 ambient;
    vec3 diffuse;
    vec3 specular;
};

struct PointLight {
    vec3 position;
    
    float constant;
    float linear;
    float quadratic;
	
    vec3 ambient;
    vec3 diffuse;
    vec3 specular;
};

in vec4 EyeSpacePos;
struct NormalMapCalc {
	//for normal map calculation
	vec3 TangentLightPos[NR_POINT_LIGHTS];
	vec3 TangentViewPos;
	vec3 TangentFragPos;
	vec3 TangentNormal;
	vec3 TangentLightDiretion;
};

in vec3 FragPos;
in vec3 Normal;
in vec2 TexCoords;
in mat3 TBN_fromTangentTWorld;;
in vec3 viewPos_out;

in DirLight dirLight_out;
in PointLight pointLights_out[NR_POINT_LIGHTS];
in NormalMapCalc normalMapCalc_out;

uniform bool fog;
const vec4 fogColor = vec4(0.5, 0.5, 0.5, 1.0);
const float fogDensity = 0.25;

uniform sampler2D texture_diffuse1;
uniform sampler2D texture_normal1;
uniform sampler2D texture_specular1; //in land no specular map exists
uniform float shininess = 32.0;

uniform bool NormalMappingEnabled; //states if normalmapping is enabled


// functions
vec3 CalcDirLight_inTangentSpace(DirLight light, NormalMapCalc normalMapCalc);
vec3 CalcPointLight_inTangentSpace(PointLight light, NormalMapCalc normalMapCalc, vec3 fragPos, int currentPointLight);
vec3 CalcDirLight(DirLight light, vec3 normal, vec3 viewDir);
vec3 CalcPointLight(PointLight light, vec3 normal, vec3 fragPos, vec3 viewDir);
float getFogFactor(float fDensity);

void main()
{   
	//get normal from normal map
	vec3 normal = texture(texture_normal1, TexCoords).rgb;
	//transform normal vector from range [-1,1] to [0,1]
	normal = normalize(normal * 2.0 - 1.0); 
	
	vec3 result;

	if(NormalMappingEnabled)
	{
		NormalMapCalc normalMapCalc = normalMapCalc_out;
	    normalMapCalc.TangentNormal = normal;
		result = CalcDirLight_inTangentSpace(dirLight_out, normalMapCalc);

		// phase 2: point lights
		for(int i = 0; i < NR_POINT_LIGHTS; i++)
		{
			result += CalcPointLight_inTangentSpace(pointLights_out[i], normalMapCalc, FragPos, i);	
		}		
    
	}else
	{
		// NOT in tangent space
		//vec3 normal = normalize(Normal); // in case no normal map is attached
		vec3 Norm = normalize(normal);
		vec3 viewDir = normalize(viewPos_out - FragPos);
		
		// Calculate and add all lights up
		// phase 1: directional lighting
		result = CalcDirLight(dirLight_out, Norm, viewDir);

		 // phase 2: point lights
		for(int i = 0; i < NR_POINT_LIGHTS; i++)
		{
			result += CalcPointLight(pointLights_out[i], Norm, FragPos, viewDir);   
		}	   
	}
	
    FragColor = vec4(result, 1.0);

	// Add fog
	//mix function: (1.0 - fogFactor)*fogColor + fogFactor * FragColor
	if(fog){
		FragColor = mix(fogColor, FragColor, getFogFactor(fogDensity));
		
		//!!!!! OpenGl Superbible Solution !!!!!!!
		//float z = length(EyeSpacePos);
		//float de = 0.25 * smoothstep(0.0, 6.0, 10.0 - FragPos.y);
		//float di = 0.45 * smoothstep(0.0, 40.0, 20.0 - FragPos.y);

		//float extinction = exp(-z * de);
		//float inscattering = exp(-z * di);
		//FragColor = vec4(result * extinction + vec3(0.5, 0.5, 0.5) * (1.0 - inscattering), 1.0);
	}
	

}

//########## Additional functions ##########

vec3 CalcDirLight_inTangentSpace(DirLight light, NormalMapCalc normalMapCalc)
{
	vec3 TangentLightDir = normalize(-normalMapCalc.TangentLightDiretion);
	
	// diffuse shading
	float diff = max(dot(normalMapCalc.TangentNormal, TangentLightDir), 0.0);
	
	// specular shading
	vec3 reflectDir = reflect(-TangentLightDir, normalMapCalc.TangentNormal);
	vec3 viewDir = normalize(normalMapCalc.TangentViewPos - normalMapCalc.TangentFragPos);
    float spec = pow(max(dot(viewDir, reflectDir), 0.0), shininess);

	 // combine results
	vec3 ambient = light.ambient * vec3(texture(texture_diffuse1, TexCoords));
    vec3 diffuse = light.diffuse * diff * vec3(texture(texture_diffuse1, TexCoords));
	vec3 specular = light.specular * spec * vec3(texture(texture_specular1, TexCoords));
	//vec3 specular = light.specular * spec * vec3(texture(texture_diffuse1, TexCoords));
	
	return (ambient + diffuse + specular);
}

// calculates the color when using a directional light.
vec3 CalcDirLight(DirLight light, vec3 normal, vec3 viewDir)
{
    vec3 lightDir = normalize(-light.direction);
    
	// diffuse shading
    float diff = max(dot(normal, lightDir), 0.0);
    
	// specular shading
    vec3 reflectDir = reflect(-lightDir, normal);
    float spec = pow(max(dot(viewDir, reflectDir), 0.0), shininess);
    
	// combine results
    vec3 ambient = light.ambient * vec3(texture(texture_diffuse1, TexCoords));
    vec3 diffuse = light.diffuse * diff * vec3(texture(texture_diffuse1, TexCoords));
    vec3 specular = light.specular * spec * vec3(texture(texture_specular1, TexCoords));
	//vec3 specular = light.specular * spec * vec3(texture(texture_diffuse1, TexCoords));
    
	return (ambient + diffuse + specular);
}

// calculates the color when using a point light.
vec3 CalcPointLight(PointLight light, vec3 normal, vec3 fragPos, vec3 viewDir)
{
    vec3 lightDir = normalize(light.position - fragPos);
    
	// diffuse shading
    float diff = max(dot(normal, lightDir), 0.0);
    
	// specular shading
    vec3 reflectDir = reflect(-lightDir, normal);
    float spec = pow(max(dot(viewDir, reflectDir), 0.0), shininess);
    
	// attenuation
    float distance = length(light.position - fragPos);
    float attenuation = 1.0 / (light.constant + light.linear * distance + light.quadratic * (distance * distance));    
    
	// combine results
    vec3 ambient = light.ambient * vec3(texture(texture_diffuse1, TexCoords));
    vec3 diffuse = light.diffuse * diff * vec3(texture(texture_diffuse1, TexCoords));
    vec3 specular = light.specular * spec * vec3(texture(texture_specular1, TexCoords));
	//vec3 specular = light.specular * spec * vec3(texture(texture_diffuse1, TexCoords));
    
	ambient *= attenuation;
    diffuse *= attenuation;
    specular *= attenuation;
    
	return (ambient + diffuse + specular);
}

// calculates the color when using a point light.
vec3 CalcPointLight_inTangentSpace(PointLight light, NormalMapCalc normalMapCalc, vec3 fragPos, int currentPointLight)
{
	vec3 lightDir = normalize(normalMapCalc.TangentLightPos[currentPointLight] - normalMapCalc.TangentFragPos);
	
	// diffuse shading
    float diff = max(dot(normalMapCalc.TangentNormal, lightDir), 0.0);
    
	// specular shading
    vec3 reflectDir = reflect(-lightDir, normalMapCalc.TangentNormal);
	vec3 TangentviewDir = normalize(normalMapCalc.TangentViewPos - normalMapCalc.TangentFragPos);
    float spec = pow(max(dot(TangentviewDir, reflectDir), 0.0), shininess);
    
	// attenuation
    float distance = length(normalMapCalc.TangentLightPos[currentPointLight] -  normalMapCalc.TangentFragPos);
    float attenuation = 1.0 / (light.constant + light.linear * distance + light.quadratic * (distance * distance));    
    
	// combine results
    vec3 ambient = light.ambient * vec3(texture(texture_diffuse1, TexCoords));
    vec3 diffuse = light.diffuse * diff * vec3(texture(texture_diffuse1, TexCoords));
    vec3 specular = light.specular * spec * vec3(texture(texture_specular1, TexCoords));
	//vec3 specular = light.specular * spec * vec3(texture(texture_diffuse1, TexCoords));
    
	ambient *= attenuation;
    diffuse *= attenuation;
    specular *= attenuation;
    
	return (ambient + diffuse + specular);
}

//Calculates the fog in the scene
float getFogFactor(float fDensity)
{
	//distance between the vertex and and view camera
	float dist = length(EyeSpacePos);
	//Exponential squared fog calculation
	float fogFactor = 1.0 /exp( (dist * fDensity) * (dist * fDensity));
	fogFactor = clamp( fogFactor, 0.0, 1.0 );

	return fogFactor;
}