#version 440 core

out vec4 FragColor;

#define NR_POINT_LIGHTS 1

struct DirLight {
    vec3 direction;
	
    vec3 ambient;
    vec3 diffuse;
    vec3 specular;
};

struct PointLight {
    vec3 position;
    
    float constant;
    float linear;
    float quadratic;
	
    vec3 ambient;
    vec3 diffuse;
    vec3 specular;
};


struct NormalMapCalc {
	//for normal map calculation
	vec3 TangentLightPos[NR_POINT_LIGHTS];
	vec3 TangentViewPos;
	vec3 TangentFragPos;
	vec3 TangentNormal;
	vec3 TangentLightDiretion;
};

in vec3 FragPos;
in vec3 Normal;
in vec2 TexCoords;

in mat3 TBN_fromTangentTWorld;;
in vec3 viewPos_out;
in mat4 modelMatrix_out;
in vec4 EyeSpacePos;

in DirLight dirLight_out;
in PointLight pointLights_out[NR_POINT_LIGHTS];
in NormalMapCalc normalMapCalc_out;

//fog
uniform bool fog;
const vec4 fogColor = vec4(0.5, 0.5, 0.5, 1.0);
uniform float fogDensity; //0.015;

//all textures
uniform sampler2D texture_diffuse1;
uniform sampler2D texture_normal1;
uniform sampler2D texture_specular1;
uniform float shininess = 32.0;

//NormalMapping
uniform bool NormalMappingEnabled; //states if normalmapping is enabled

// ShadowMapping
uniform samplerCube cubemap_tex;
uniform float near_plane; // near plane
uniform float far_plane; // far plane
uniform vec4 lightPos;

float shadow = 0;

// array of offset direction for sampling
vec3 gridSamplingDisk[20] = vec3[]
(
   vec3(1, 1,  1), vec3( 1, -1,  1), vec3(-1, -1,  1), vec3(-1, 1,  1), 
   vec3(1, 1, -1), vec3( 1, -1, -1), vec3(-1, -1, -1), vec3(-1, 1, -1),
   vec3(1, 1,  0), vec3( 1, -1,  0), vec3(-1, -1,  0), vec3(-1, 1,  0),
   vec3(1, 0,  1), vec3(-1,  0,  1), vec3( 1,  0, -1), vec3(-1, 0, -1),
   vec3(0, 1,  1), vec3( 0, -1,  1), vec3( 0, -1, -1), vec3( 0, 1, -1)
);

// functions
vec3 CalcDirLight_inTangentSpace(DirLight light, NormalMapCalc normalMapCalc);
vec3 CalcPointLight_inTangentSpace(PointLight light, NormalMapCalc normalMapCalc, vec3 fragPos, int currentPointLight);
vec3 CalcDirLight(DirLight light, vec3 normal, vec3 viewDir);
vec3 CalcPointLight(PointLight light, vec3 normal, vec3 fragPos, vec3 viewDir);
float getFogFactor(float fDensity);
float ShadowCalculation(vec3 fragPos, vec3 LightPos);

void main()
{		
	//only for trees relevant
	vec4 alpha = texture(texture_diffuse1,TexCoords).aaaa;
	if(alpha.a < 0.1)
	{
		discard;
	}

	//get normal from normal map
	vec3 normal = texture(texture_normal1, TexCoords).rgb;
	//transform normal vector from range [-1,1] to [0,1]
	normal = normalize(normal * 2.0 - 1.0); 
	
	vec3 result;

	// calculate shadow
    shadow = ShadowCalculation(FragPos, vec3(lightPos));  

	if(NormalMappingEnabled)
	{
		NormalMapCalc normalMapCalc = normalMapCalc_out;
	    normalMapCalc.TangentNormal = normal;
		result = CalcDirLight_inTangentSpace(dirLight_out, normalMapCalc);

		// phase 2: point lights
		for(int i = 0; i < NR_POINT_LIGHTS; i++)
		{
			result += CalcPointLight_inTangentSpace(pointLights_out[i], normalMapCalc, FragPos, i);	
		}		
    
	}
	else
	{
		// NOT in tangent space
		normal = normalize(Normal); // in case no normal map is attached
		vec3 Norm = normal;
		vec3 viewDir = normalize(viewPos_out - FragPos);
		
		// Calculate and add all lights up
		// phase 1: directional lighting
		result = CalcDirLight(dirLight_out, Norm, viewDir);

		 // phase 2: point lights
		for(int i = 0; i < NR_POINT_LIGHTS; i++)
		{
			result += CalcPointLight(pointLights_out[i], Norm, FragPos, viewDir);   
		}	   
	}            

    FragColor = vec4(result, 1.0);
	//FragColor = vec4(vec3(shadow / far_plane), 1.0);  //to test depth map

	// Add fog
	if(fog){
		//mix function: (1.0 - fogFactor)*fogColor + fogFactor * FragColor
		FragColor = vec4(mix(fogColor, FragColor, getFogFactor(fogDensity)).rgb,1.0);
	}
}

//########## Additional functions ##########

vec3 CalcDirLight_inTangentSpace(DirLight light, NormalMapCalc normalMapCalc)
{
	vec3 TangentLightDir = normalize(-normalMapCalc.TangentLightDiretion);
	
	// diffuse shading
	float diff = max(dot(normalMapCalc.TangentNormal, TangentLightDir), 0.0);
	
	// specular shading
	vec3 reflectDir = reflect(-TangentLightDir, normalMapCalc.TangentNormal);
	vec3 viewDir = normalize(normalMapCalc.TangentViewPos - normalMapCalc.TangentFragPos);
    float spec = pow(max(dot(viewDir, reflectDir), 0.0), shininess);

	 // combine results
	vec3 ambient = light.ambient * vec3(texture(texture_diffuse1, TexCoords));
    vec3 diffuse = light.diffuse * diff * vec3(texture(texture_diffuse1, TexCoords));
	vec3 specular = light.specular * spec * vec3(texture(texture_specular1, TexCoords));
	//vec3 specular = light.specular * spec * vec3(texture(texture_diffuse1, TexCoords));
	
	//return (ambient +  (1.0 - shadow) * (diffuse + specular));
	return (ambient +  diffuse + specular);
}

// calculates the color when using a directional light.
vec3 CalcDirLight(DirLight light, vec3 normal, vec3 viewDir)
{
    vec3 lightDir = normalize(-light.direction);
    
	// diffuse shading
    float diff = max(dot(normal, lightDir), 0.0);
    
	// specular shading
    vec3 reflectDir = reflect(-lightDir, normal);
    float spec = pow(max(dot(viewDir, reflectDir), 0.0), shininess);
    
	// combine results
    vec3 ambient = light.ambient * vec3(texture(texture_diffuse1, TexCoords));
    vec3 diffuse = light.diffuse * diff * vec3(texture(texture_diffuse1, TexCoords));
    vec3 specular = light.specular * spec * vec3(texture(texture_specular1, TexCoords));
	//vec3 specular = light.specular * spec * vec3(texture(texture_diffuse1, TexCoords));
    
	//return (ambient +  (1.0 - shadow) * (diffuse + specular));
	return (ambient +  diffuse + specular);
}

// calculates the color when using a point light.
vec3 CalcPointLight(PointLight light, vec3 normal, vec3 fragPos, vec3 viewDir)
{
    vec3 lightDir = normalize(light.position - fragPos);
    
	// diffuse shading
    float diff = max(dot(normal, lightDir), 0.0);
    
	// specular shading
    vec3 reflectDir = reflect(-lightDir, normal);
    float spec = pow(max(dot(viewDir, reflectDir), 0.0), shininess);
    
	// attenuation
    float distance = (length(light.position - fragPos));
    float attenuation = 1.0 / (light.constant + light.linear * distance + light.quadratic * (distance * distance));    
    
	// combine results
    vec3 ambient = light.ambient * vec3(texture(texture_diffuse1, TexCoords));
    vec3 diffuse = light.diffuse * diff * vec3(texture(texture_diffuse1, TexCoords));
    vec3 specular = light.specular * spec * vec3(texture(texture_specular1, TexCoords));
	//vec3 specular = light.specular * spec * vec3(texture(texture_diffuse1, TexCoords));
    
	ambient *= attenuation;
    diffuse *= attenuation;
    specular *= attenuation;
    
	return (ambient +  (1.0 - shadow) * (diffuse + specular));
}

// calculates the color when using a point light.
vec3 CalcPointLight_inTangentSpace(PointLight light, NormalMapCalc normalMapCalc, vec3 fragPos, int currentPointLight)
{
	vec3 lightDir = normalize(normalMapCalc.TangentLightPos[currentPointLight] - normalMapCalc.TangentFragPos);
	
	// diffuse shading
    float diff = max(dot(normalMapCalc.TangentNormal, lightDir), 0.0);
    
	// specular shading
    vec3 reflectDir = reflect(-lightDir, normalMapCalc.TangentNormal);
	vec3 TangentviewDir = normalize(normalMapCalc.TangentViewPos - normalMapCalc.TangentFragPos);
    float spec = pow(max(dot(TangentviewDir, reflectDir), 0.0), shininess);
    
	// attenuation
    float distance = length(normalMapCalc.TangentLightPos[currentPointLight] -  normalMapCalc.TangentFragPos);
    float attenuation = 1.0 / (light.constant + light.linear * distance + light.quadratic * (distance * distance));    
    
	// combine results
    vec3 ambient = light.ambient * vec3(texture(texture_diffuse1, TexCoords));
    vec3 diffuse = light.diffuse * diff * vec3(texture(texture_diffuse1, TexCoords));
    //vec3 specular = light.specular * spec * vec3(texture(texture_specular1, TexCoords));
	vec3 specular = light.specular * spec * vec3(texture(texture_diffuse1, TexCoords));
    
	ambient *= attenuation;
    diffuse *= attenuation;
    specular *= attenuation;
    
	return (ambient +  (1.0 - shadow) * (diffuse + specular));
}
 
//Calculates the fog in the scene
float getFogFactor(float fDensity)
{
	//distance between the vertex and and view camera
	float dist = length(EyeSpacePos);
	//Exponential squared fog calculation
	float fogFactor = 1.0 /exp( (dist * fDensity) * (dist * fDensity));
	fogFactor = clamp( fogFactor, 0.0, 1.0 );

	return fogFactor;
}

float ShadowCalculation(vec3 fragPos, vec3 LightPos)
{
    // get vector between fragment position and light position
    vec3 fragToLight = fragPos - LightPos;
    // use the light to fragment vector to sample from the depth map    
   // float closestDepth = texture(cubemap_tex, fragToLight).r;
    // it is currently in linear range between [0,1]. Re-transform back to original value
   // closestDepth *= far_plane;
    // now get current linear depth as the length between the fragment and light position
    float currentDepth = length(fragToLight);
  
   //PCF algorithm
   float bias = 0.001;
   int samples = 20;
   float shadow = 0.0f;

	float viewDistance = length(viewPos_out - fragPos);
	float diskRadius = (1.0 + (viewDistance / far_plane)) / 50.0;
	
	float closestDepth=0;

	for(int i = 0; i < samples; ++i)
	{
		closestDepth = texture(cubemap_tex, fragToLight + gridSamplingDisk[i] * diskRadius).r;
		closestDepth *= far_plane;   // undo mapping [0;1]
		if(currentDepth - bias > closestDepth)
			shadow += 1.0;
	}

		shadow /= float(samples);
	
    return shadow;
    //return closestDepth; //to test depth map
}  