#version 440 core
out vec4 FragColor;

struct DirLight {
    vec3 direction;
	
    vec3 ambient;
    vec3 diffuse;
    vec3 specular;
};

struct PointLight {
    vec3 position;
    
    float constant;
    float linear;
    float quadratic;
	
    vec3 ambient;
    vec3 diffuse;
    vec3 specular;
};

in vec4 EyeSpacePos;
in vec3 FragPos;
in vec3 Normal;
in vec2 TexCoords;

uniform bool fog;
const vec4 fogColor = vec4(0.5, 0.5, 0.5, 1.0);
const float fogDensity = 0.25;

uniform sampler2D texture_diffuse1;
uniform sampler2D texture_specular1;
uniform float shininess = 32.0;

#define NR_POINT_LIGHTS 1
uniform vec3 viewPos;
uniform DirLight dirLight;
uniform PointLight pointLights[NR_POINT_LIGHTS];

// functions
vec3 CalcDirLight(DirLight light, vec3 normal, vec3 viewDir);
vec3 CalcPointLight(PointLight light, vec3 normal, vec3 fragPos, vec3 viewDir);
float getFogFactor(float fDensity);

void main()
{    
	vec4 alpha = texture(texture_diffuse1,TexCoords).aaaa;
	if(alpha.a < 0.1)
	{
		discard;
	}

	// properties
    vec3 norm = normalize(Normal);
    vec3 viewDir = normalize(viewPos - FragPos);

	// Calculate and add all lights up
    //1: directional lighting
   vec3 result = CalcDirLight(dirLight, norm, viewDir);
    //2: point lights
   for(int i = 0; i < NR_POINT_LIGHTS; i++)
		result += CalcPointLight(pointLights[i], norm, FragPos, viewDir);      
    
	//if texture has an alpha value smaller than 0.1
	//then discard the pixels --> used for leaves

	FragColor = vec4(result, 1.0);

	// Add fog
	//mix function: (1.0 - fogFactor)*fogColor + fogFactor * FragColor
	if(fog){
		FragColor = mix(fogColor, FragColor, getFogFactor(fogDensity));
	}

}

// calculates the color when using a directional light.
vec3 CalcDirLight(DirLight light, vec3 normal, vec3 viewDir)
{
    vec3 lightDir = normalize(-light.direction);
    // diffuse shading
    float diff = max(dot(normal, lightDir), 0.0);
    // specular shading
    vec3 reflectDir = reflect(-lightDir, normal);
    float spec = pow(max(dot(viewDir, reflectDir), 0.0), shininess);
    // combine results
    vec3 ambient = light.ambient * vec3(texture(texture_diffuse1, TexCoords));
    vec3 diffuse = light.diffuse * diff * vec3(texture(texture_diffuse1, TexCoords));
    //vec3 specular = light.specular * spec * vec3(texture(texture_specular1, TexCoords));
	vec3 specular = light.specular * spec * vec3(texture(texture_diffuse1, TexCoords));
    return (ambient + diffuse + specular);
}


// calculates the color when using a point light.
vec3 CalcPointLight(PointLight light, vec3 normal, vec3 fragPos, vec3 viewDir)
{
    vec3 lightDir = normalize(light.position - fragPos);
    // diffuse shading
    float diff = max(dot(normal, lightDir), 0.0);
    // specular shading
    vec3 reflectDir = reflect(-lightDir, normal);
    float spec = pow(max(dot(viewDir, reflectDir), 0.0), shininess);
    // attenuation
    float distance = length(light.position - fragPos);
    float attenuation = 1.0 / (light.constant + light.linear * distance + light.quadratic * (distance * distance));    
    // combine results
    vec3 ambient = light.ambient * vec3(texture(texture_diffuse1, TexCoords));
    vec3 diffuse = light.diffuse * diff * vec3(texture(texture_diffuse1, TexCoords));
    vec3 specular = light.specular * spec * vec3(texture(texture_diffuse1, TexCoords));
	//vec3 specular = light.specular * spec * vec3(texture(texture_specular1, TexCoords));
    ambient *= attenuation;
    diffuse *= attenuation;
    specular *= attenuation;
    return (ambient + diffuse + specular);
}

//Calculates the fog in the scene
float getFogFactor(float fDensity)
{
	//distance between the vertex and and view camera
	float dist = length(EyeSpacePos);
	//Exponential squared fog calculation
	float fogFactor = 1.0 /exp( (dist * fDensity) * (dist * fDensity));
	fogFactor = clamp( fogFactor, 0.0, 1.0 );

	return fogFactor;
}